@@ -43,6 +43,16 @@ typedef struct DirectoryChildData DirectoryChildData;
 typedef struct DirectoryProcessingData DirectoryProcessingData;
 typedef struct DirectoryRootInfo DirectoryRootInfo;
 
+typedef struct {
+	TrackerCrawler *crawler;
+	TrackerEnumerator *enumerator;
+	DirectoryRootInfo  *root_info;
+	DirectoryProcessingData *dir_info;
+	GFile *dir_file;
+	GCancellable *cancellable;
+	GSList *files;
+} DataProviderData;
+
 struct DirectoryChildData {
 	GFile          *child;
 	gboolean        is_dir;
@@ -62,6 +72,10 @@ struct DirectoryRootInfo {
 
 	GQueue *directory_processing_queue;
 
+	TrackerDirectoryFlags flags;
+
+	DataProviderData *dpd;
+
 	/* Directory stats */
 	guint directories_found;
 	guint directories_ignored;
@@ -94,16 +108,6 @@ struct TrackerCrawlerPrivate {
 	gboolean        was_started;
 };
 
-typedef struct {
-	TrackerCrawler *crawler;
-	TrackerEnumerator *enumerator;
-	DirectoryRootInfo  *root_info;
-	DirectoryProcessingData *dir_info;
-	GFile *dir_file;
-	GCancellable *cancellable;
-	GSList *files;
-} DataProviderData;
-
 enum {
 	CHECK_DIRECTORY,
 	CHECK_FILE,
@@ -132,10 +136,13 @@ static gboolean check_defaults           (TrackerCrawler  *crawler,
 static gboolean check_contents_defaults  (TrackerCrawler  *crawler,
                                           GFile           *file,
                                           GList           *contents);
+static void     data_provider_data_free  (DataProviderData        *dpd);
+
 static void     data_provider_begin      (TrackerCrawler          *crawler,
 					  DirectoryRootInfo       *info,
 					  DirectoryProcessingData *dir_data);
-
+static void     data_provider_end        (TrackerCrawler          *crawler,
+                                          DirectoryRootInfo       *info);
 static void     directory_root_info_free (DirectoryRootInfo *info);
 
 
@@ -459,14 +466,14 @@ directory_processing_data_add_child (DirectoryProcessingData *data,
 }
 
 static DirectoryRootInfo *
-directory_root_info_new (GFile             *file,
-                         gint               max_depth,
-                         gchar             *file_attributes,
-                         TrackerCrawlFlags  flags)
+directory_root_info_new (GFile                 *file,
+                         gint                   max_depth,
+                         gchar                 *file_attributes,
+                         TrackerDirectoryFlags  flags)
 {
 	DirectoryRootInfo *info;
 	DirectoryProcessingData *dir_info;
-	gboolean enable_stat;
+	gboolean allow_stat = TRUE;
 
 	info = g_slice_new0 (DirectoryRootInfo);
 
@@ -476,14 +483,21 @@ directory_root_info_new (GFile             *file,
 
 	info->tree = g_node_new (g_object_ref (file));
 
-	enable_stat = (flags & TRACKER_CRAWL_FLAG_NO_STAT) == 0;
+	info->flags = flags;
 
-	if (enable_stat && file_attributes) {
+	if ((info->flags & TRACKER_DIRECTORY_FLAG_NO_STAT) != 0) {
+		allow_stat = FALSE;
+	}
+
+	if (allow_stat && file_attributes) {
 		GFileInfo *file_info;
+		GFileQueryInfoFlags file_flags;
+
+		file_flags = G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS;
 
 		file_info = g_file_query_info (file,
 		                               file_attributes,
-		                               G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+		                               file_flags,
 		                               NULL,
 		                               NULL);
 		g_object_set_qdata_full (G_OBJECT (file),
@@ -510,6 +524,11 @@ directory_tree_free_foreach (GNode    *node,
 static void
 directory_root_info_free (DirectoryRootInfo *info)
 {
+	if (info->dpd)  {
+		data_provider_data_free (info->dpd);
+		info->dpd = NULL;
+	}
+
 	g_object_unref (info->directory);
 
 	g_node_traverse (info->tree,
@@ -624,6 +643,7 @@ process_func (gpointer data)
 			       info->files_found,
 			       info->files_ignored);
 
+		data_provider_end (crawler, info);
 		g_queue_pop_head (priv->directories);
 		directory_root_info_free (info);
 	}
@@ -781,6 +801,79 @@ data_provider_data_free (DataProviderData *dpd)
 	g_slice_free (DataProviderData, dpd);
 }
 
+static void
+data_provider_end_cb (GObject      *object,
+                      GAsyncResult *result,
+                      gpointer      user_data)
+{
+	DataProviderData *dpd;
+	GError *error = NULL;
+	gboolean cancelled;
+
+	dpd = user_data;
+
+	cancelled = g_cancellable_is_cancelled (dpd->cancellable);
+
+	tracker_data_provider_end_finish (TRACKER_DATA_PROVIDER (object), result, &error);
+
+	if (error) {
+		if (!cancelled) {
+			GFile *parent;
+			gchar *uri;
+
+			parent = dpd->dir_info->node->data;
+			uri = g_file_get_uri (parent);
+
+			g_warning ("Could not end data provider for container / directory '%s', %s",
+			           uri, error ? error->message : "no error given");
+
+			g_free (uri);
+		}
+
+		g_clear_error (&error);
+	}
+
+	data_provider_data_free (dpd);
+}
+
+static void
+data_provider_end (TrackerCrawler    *crawler,
+                   DirectoryRootInfo *info)
+{
+	DataProviderData *dpd;
+
+	g_return_if_fail (info != NULL);
+
+	if (info->dpd == NULL) {
+		/* Nothing to do */
+		return;
+	}
+
+	/* We detach the DataProviderData from the DirectoryRootInfo
+	 * here so it's not freed early. We can't use
+	 * DirectoryRootInfo as user data for the async function below
+	 * because it's freed before that callback will be called.
+	 */
+	dpd = info->dpd;
+	info->dpd = NULL;
+
+	if (dpd->enumerator) {
+		/* Reset cancellation, we need to perform this even
+		 * if we were cancelled previously.
+		 */
+		g_cancellable_reset (dpd->cancellable);
+
+		tracker_data_provider_end_async (crawler->priv->data_provider,
+		                                 dpd->enumerator,
+		                                 G_PRIORITY_LOW,
+		                                 dpd->cancellable,
+		                                 data_provider_end_cb,
+		                                 dpd);
+	} else {
+		data_provider_data_free (dpd);
+	}
+}
+
 static void
 enumerate_next_cb (GObject      *object,
                    GAsyncResult *result,
@@ -797,11 +890,11 @@ enumerate_next_cb (GObject      *object,
 
 	/* If cancelled, process what we have so far only... */
 	if (cancelled) {
-		data_provider_data_process (dpd);
 		data_provider_data_add (dpd);
+		data_provider_data_process (dpd);
 
 		process_func_start (dpd->crawler);
-		data_provider_data_free (dpd);
+
 		return;
 	}
 
@@ -834,12 +927,11 @@ enumerate_next_cb (GObject      *object,
 		} else {
 			/* condition c) */
 			/* Done enumerating, start processing what we got ... */
-			data_provider_data_process (dpd);
 			data_provider_data_add (dpd);
+			data_provider_data_process (dpd);
 		}
 
 		process_func_start (dpd->crawler);
-		data_provider_data_free (dpd);
 	} else {
 		/* More work to do, we keep reference given to us */
 		dpd->files = g_slist_prepend (dpd->files, info);
@@ -857,11 +949,14 @@ data_provider_begin_cb (GObject      *object,
                         GAsyncResult *result,
                         gpointer      user_data)
 {
+	DirectoryRootInfo *info;
 	DataProviderData *dpd;
 	GError *error = NULL;
 	gboolean cancelled;
 
-	dpd = user_data;
+	info = user_data;
+	dpd = info->dpd;
+
 	cancelled = g_cancellable_is_cancelled (dpd->cancellable);
 
 	dpd->enumerator = tracker_data_provider_begin_finish (TRACKER_DATA_PROVIDER (object), result, &error);
@@ -882,13 +977,13 @@ data_provider_begin_cb (GObject      *object,
 		}
 
 		process_func_start (dpd->crawler);
-		data_provider_data_free (dpd);
+
 		return;
 	}
 
 	if (cancelled) {
 		process_func_start (dpd->crawler);
-		data_provider_data_free (dpd);
+
 		return;
 	}
 
@@ -907,7 +1002,13 @@ data_provider_begin (TrackerCrawler          *crawler,
 	DataProviderData *dpd;
 	gchar *attrs;
 
+	/* DataProviderData is freed in data_provider_end() call. This
+	 * call must _ALWAYS_ be reached even on cancellation or
+	 * failure, this is normally the case when we return to the
+	 * process_func() and finish a directory.
+	 */
 	dpd = data_provider_data_new (crawler, info, dir_data);
+	info->dpd = dpd;
 
 	if (crawler->priv->file_attributes) {
 		attrs = g_strconcat (FILE_ATTRIBUTES ",",
@@ -917,25 +1018,24 @@ data_provider_begin (TrackerCrawler          *crawler,
 		attrs = g_strdup (FILE_ATTRIBUTES);
 	}
 
-
 	tracker_data_provider_begin_async (crawler->priv->data_provider,
 	                                   dpd->dir_file,
 	                                   attrs,
-	                                   G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+	                                   info->flags,
 	                                   G_PRIORITY_LOW,
 	                                   dpd->cancellable,
 	                                   data_provider_begin_cb,
-	                                   dpd);
+	                                   info);
 	g_free (attrs);
 }
 
 gboolean
-tracker_crawler_start (TrackerCrawler *crawler,
-                       GFile          *file,
-                       gint            max_depth)
+tracker_crawler_start (TrackerCrawler        *crawler,
+                       GFile                 *file,
+                       TrackerDirectoryFlags  flags,
+                       gint                   max_depth)
 {
 	TrackerCrawlerPrivate *priv;
-	TrackerCrawlFlags flags;
 	DirectoryRootInfo *info;
 	gboolean enable_stat;
 
@@ -944,8 +1044,7 @@ tracker_crawler_start (TrackerCrawler *crawler,
 
 	priv = crawler->priv;
 
-	flags = tracker_data_provider_get_crawl_flags (priv->data_provider);
-	enable_stat = (flags & TRACKER_CRAWL_FLAG_NO_STAT) == 0;
+	enable_stat = (flags & TRACKER_DIRECTORY_FLAG_NO_STAT) == 0;
 
 	if (enable_stat && !g_file_query_exists (file, NULL)) {
 		/* This shouldn't happen, unless the removal/unmount notification
